import argparse
import sys

from kmpc.version import VERSION_STR

parser = argparse.ArgumentParser()
parser.add_argument("-q","--quiet",help="only print errors to console log",action="store_true")
parser.add_argument("-d","--debug",help="print debug messages to console log",action="store_true")
parser.add_argument("-n","--newconfig",help="write out default config file if it doesn't exist yet",action="store_true")
parser.add_argument("-V","--version",help="print version number and exit",action="store_true")
parser.add_argument("--helpkivy",help="Print Kivy's built-in argument list",action="store_true")

def do_args():
    # since kivy has it's own argparsing, it's necessary to do some argv mangling
    args,unknown = parser.parse_known_args()
    sys.argv[1:] = unknown
    # if --helpkivy is passed, print Kivy's argument list
    if args.helpkivy:
        sys.argv.append('-h')
    # if -d/--debug is passed, use Kivy's -d flag
    if args.debug:
        sys.argv.append('-d')
    # if -n/--newconfig is passed, write config and exit
    if args.newconfig:
        print "Writing default config file if it does not exist"
        print ""
        from kmpc.extra import KmpcHelpers
        Helpers=KmpcHelpers()
        Helpers.loadconfigfile()
        sys.exit(0)
    if args.version:
        print sys.argv[0]+" v"+VERSION_STR
        sys.exit(0)
    from kivy.config import Config
    if args.quiet:
        Config.set('kivy','log_level','warning')
    return args

def main_app():
    args=do_args()
    from kmpc import kmpcapp
    kmpcapp.KmpcApp(args).run()

def manager_app():
    args=do_args()
    from kmpc import kmpcmanager
    kmpcmanager.ManagerApp(args).run()



import os
import ConfigParser
from kmpc.mpd import MPDProtocol
from kmpc.mpdfactory import MPDClientFactory

import kivy
kivy.require('1.10.0')

#install twisted reactor to interface with mpd
from kivy.support import install_twisted_reactor
install_twisted_reactor()
from twisted.internet import reactor, protocol
from twisted.internet.defer import inlineCallbacks

from kivy.app import App
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.slider import Slider
from kivy.uix.label import Label
from kivy.uix.tabbedpanel import TabbedPanelItem

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

# this class just returns a debug message for all calls to it to handle bad mpd connections
class Dummy(object):
    def __getattr__(self,attr):
        Logger.debug("MpdConnection: no connection when calling "+attr+" method")
        return self
    def __call__(self,*args):
        return self

class MpdConnection(object):

    def __init__(self,config,idlehandler=None,initconnections=[]):
        self.config = config
        # set up mpd connection
        self.initconnections=initconnections
        self.factory = MPDClientFactory(idlehandler)
        self.factory.connectionMade = self.mpd_connectionMade
        self.factory.connectionLost = self.mpd_connectionLost
        reactor.connectTCP(self.config.get('mpd','mpdhost'), self.config.getint('mpd','mpdport'), self.factory)
        self.noprotocol=Dummy()

    # this part handles calls to protocol when it hasn't been set up yet or is incorrectly specified in config
    @property
    def protocol(self):
        try:
            if self.realprotocol:
                return self.realprotocol
        except AttributeError:
            Logger.debug("MpdConnection: no mpd connected")
            return self.noprotocol

    def mpd_connectionMade(self,protocol):
        """Callback when mpd is connected."""
        # copy the protocol to all the classes
        self.realprotocol = protocol
        Logger.info('Application: Connected to mpd server host='+self.config.get('mpd','mpdhost')+' port='+self.config.get('mpd','mpdport'))
        for ic in self.initconnections:
            if callable(ic):
                 ic(self)

    def mpd_connectionLost(self,protocol, reason):
        """Callback when mpd connection is lost."""
        Logger.warn('Application: Connection lost: %s' % reason)
        # kills the app for now since I don't know how to handle this yet
        App.get_running_app().stop()

    def handle_mpd_error(self,result):
        """Prints handled errors to the error log."""
        Logger.error('Application: MPDIdleHandler Callback error: {}'.format(result))

class KmpcHelpers(object):

    def formatsong(self,rec):
        """Method used by library browser to properly format a song row."""
        song = ''
        # check if there is more than one disc and display if so
        (d1,d2)=rec['disc'].split('/')
        if int(d2) > 1:
            song+='(Disc '+'%02d' % int(d1)+') '
        # sometimes track numbers are like '01/05' (one of five), so drop that second number
        (t1,t2)=rec['track'].split('/')
        song+='%02d' % int(t1)+' '
        # if albumartist is different than track artist, display the track artist
        if rec['artist'] != rec['albumartist']:
            song+=rec['artist']+' - '
        # display the track title
        song+=rec['title']
        return song

    def loadconfigfile(self):
        # set up config with default values
        config=ConfigParser.SafeConfigParser()
        config.add_section('mpd')
        config.set('mpd','mpdhost','127.0.0.1')
        config.set('mpd','mpdport','6600')
        config.add_section('paths')
        config.set('paths','musicpath','/mnt/music')
        config.set('paths','fanartpath','/mnt/fanart')
        config.set('paths','tmppath','/tmp')
        config.add_section('sync')
        config.set('sync','synchost','127.0.0.1')
        config.set('sync','syncmusicpath','/mnt/music')
        config.set('sync','syncfanartpath','/mnt/fanart')
        config.set('sync','synctmppath','/tmp')
        config.add_section('flags')
        config.set('flags','rpienable','False')
        config.set('flags','originalyear','True')
        config.add_section('logs')
        config.set('logs','artlog','False')
        config.add_section('fanart')
        config.set('fanart','client_key','')
        config.add_section('songratings')
        config.set('songratings','zero','Silence')
        config.set('songratings','one','Songs that should never be heard')
        config.set('songratings','two','Songs no one likes')
        config.set('songratings','three','Songs for certain occasions')
        config.set('songratings','four','Songs someone else likes')
        config.set('songratings','five','Filler tracks with no music')
        config.set('songratings','six','Meh track or short musical filler')
        config.set('songratings','seven','Occasional listening songs')
        config.set('songratings','eight','Great songs for all occasions')
        config.set('songratings','nine','Best songs by an artist')
        config.set('songratings','ten','Favorite songs of all time')
        # check if config folder exists
        if os.path.isdir(configdir):
            # try to read existing config file
            config.read([os.path.join(configdir,'config.ini')])
            # write out config file in case it doesn't exist yet
            with open(os.path.join(configdir,'config.ini'),'wb') as cf:
                config.write(cf)
        else:
            os.mkdir(configdir)
            # write out config file
            with open(os.path.join(configdir,'config.ini'),'wb') as cf:
                config.write(cf)
        # return the generated config
        return config

    # fontawesome strings and subjective interpretations of song ratings.
    def songratings(self,config):
        sr= {
            '0': {'stars': u"\uf006\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','zero')},
            '1': {'stars': u"\uf123\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','one')},
            '2': {'stars': u"\uf005\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','two')},
            '3': {'stars': u"\uf005\uf123\uf006\uf006\uf006",'meaning':config.get('songratings','three')},
            '4': {'stars': u"\uf005\uf005\uf006\uf006\uf006",'meaning':config.get('songratings','four')},
            '5': {'stars': u"\uf005\uf005\uf123\uf006\uf006",'meaning':config.get('songratings','five')},
            '6': {'stars': u"\uf005\uf005\uf005\uf006\uf006",'meaning':config.get('songratings','six')},
            '7': {'stars': u"\uf005\uf005\uf005\uf123\uf006",'meaning':config.get('songratings','seven')},
            '8': {'stars': u"\uf005\uf005\uf005\uf005\uf006",'meaning':config.get('songratings','eight')},
            '9': {'stars': u"\uf005\uf005\uf005\uf005\uf123",'meaning':config.get('songratings','nine')},
            '10': {'stars': u"\uf005\uf005\uf005\uf005\uf005",'meaning':config.get('songratings','ten')},
            '' : {'stars': u"\uf29c", 'meaning':'No sticker set'}
        }
        return sr

    def getfontsize(self,str):
        """Method that determines font size based on text length."""
        # helper array for scaling font sizes based on text length
        sizearray = ['39sp','38sp','37sp','36sp','35sp','34sp','33sp','32sp','31sp','30sp','29sp','28sp','27sp','26sp','25sp']
        lr = len(str)
        if lr < 33:
            rsize = '40sp'
        elif lr >= 55:
            rsize = '24sp'
        else:
            rsize = sizearray[int(round((lr-33)/21*14))]
        return rsize

    def decodeFileName(self,name):
        """Method that tries to intelligently decode a filename to handle unicode weirdness."""
        if type(name) == str:
            try:
                name = name.decode('utf8')
            except:
                name = name.decode('windows-1252')
        return name

class ExtraSlider(Slider):
    """Class that implements some extra stuff on top of a standard slider."""

    def __init__(self,**kwargs):
        """Do normal init routine, but also register on_release event."""
        super(self.__class__,self).__init__(**kwargs)
        self.register_event_type('on_release')

    def on_release(self):
        """Override this with something you want this slider to do."""
        pass

    def on_touch_up(self, touch):
        """Check if slider is released, dispatch the on_release event if so."""
        released = super(self.__class__,self).on_touch_up(touch)
        if released:
            self.dispatch('on_release')
        return released

class OutlineLabel(Label):
    """A label that has an outline around it."""
    pass

class OutlineButton(Button,OutlineLabel):
    """A button with a label that has an outline around it."""
    pass

class ClearButton(Button,OutlineLabel):
    """A button that is clear instead of opaque."""
    pass

class OutlineTabbedPanelItem(TabbedPanelItem,OutlineLabel):
    """A label that has an outline around it."""
    pass



import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.logger import Logger
from kmpc.mpd import MPDProtocol
from twisted.internet import protocol

class MPDFactoryProtocol(MPDProtocol):
    """Factory for MPDProtocol."""
    def connectionMade(self):
        """Call the connectionMade override."""
        if callable(self.factory.connectionMade):
            self.factory.connectionMade(self)
    def connectionLost(self, reason):
        """Call the connectionLost override."""
        if callable(self.factory.connectionLost):
            self.factory.connectionLost(self, reason)

class MPDClientFactory(protocol.ClientFactory):
    """Factory for MPDClient."""
    protocol = MPDFactoryProtocol
    connectionMade = None
    connectionLost = None

    def __init__(self,idlehandler=None,**kwargs):
        self.idlehandler=idlehandler

    def buildProtocol(self, addr):
        """Hook up protocol and idle handler."""
        Logger.debug('MPDClientFactory: buildProtocol')
        protocol = self.protocol()
        protocol.factory = self
        if callable(self.idlehandler):
            protocol.idle_result = self.idlehandler
        return protocol

    def clientConnectionFailed(self, connector, reason):
        """mpd connection failed for some reason."""
        Logger.error('Connection failed - goodbye!: '+format(reason))

    def clientConnectionLost(self, connector, reason):
        """mpd connection lost for some reason."""
        Logger.error('Connection lost - goodbye!: '+format(reason))



import kivy
kivy.require('1.10.0')
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.gridlayout import GridLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.popup import Popup
from kivy.logger import Logger
from kivy.app import App
from kivy.clock import Clock
import git
import os
import stat
import codecs
from subprocess import call, PIPE, Popen
from threading import Thread
from Queue import Queue, Empty
from functools import partial

from kmpc.extra import KmpcHelpers, OutlineLabel, OutlineTabbedPanelItem
import kmpc.kmpcapp

Helpers=KmpcHelpers()

class ConfigTabbedPanelItem(OutlineTabbedPanelItem):
    """The Config tab, for internal functions."""

    def printit(self,result):
        """An internal debugging function. Probably shouldn't ever be used."""
        print format(result)

    def git_pull(self):
        """Method that runs a git pull on the current folder."""
        Logger.info('Config: git_pull')
        g = git.cmd.Git(os.getcwd())
        Logger.info(g.pull())

    def do_reboot(self):
        """Method that reboots the host."""
        Logger.info('Config: reboot')
        call(['sudo','reboot'])

    def do_poweroff(self):
        """Method that shuts down the host."""
        Logger.info('Config: poweroff')
        call(['sudo','poweroff'])

    def enqueue_output(self,out,queue,event,popup,tpath,synchost,layout,sv):
        """Method that is called in its own thread to write commandline script output to a popup, then clean up afterwards."""
        # loop through the stdout output and shove it into the queue
        for line in iter(out.readline, b''):
            queue.put(line)
        # close the output stream, cancel the .1 second update event, dismiss the popup
        out.close()
        event.cancel()
        popup.dismiss() # seems like this is too early, since we're writing more data in the next few lines
        Logger.info('Filesync: Cleaning up')
        l=OutlineLabel(text='Cleaning up temporary files',size_hint=(None,None),font_size='12sp',halign='left')
        l.bind(texture_size=l.setter('size'))
        layout.add_widget(l)
        sv.scroll_to(l)
        # try removing all temporary files, should probably catch more specific exceptions in case something truly breaks
        try:
            call(['sudo','cp',tpath,'/var/lib/mpd/playlists/root.m3u'])
            os.remove(tpath)
            os.remove(os.path.join(mainconfig.get('paths','tmppath'),'sticker.sql'))
            os.remove(os.path.join(mainconfig.get('paths','tmppath'),'scmd'))
            os.remove(os.path.join(mainconfig.get('paths','tmppath'),'sync.sh'))
            call(['ssh',synchost,'rm','-f',mainconfig.get('sync','synctmppath')+'/sticker.sql'])
            call(['ssh',synchost,'rm','-f',mainconfig.get('sync','synctmppath')+'/scmd'])
        except Exception as e:
            Logger.exception("enqueue_output: "+format(e))

    def write_queue_line(self,q,layout,sv,dt):
        """Method that checks the queue for output and adds a new label line if any output exists."""
        try: line=q.get_nowait()
        except Empty:
            pass
        else:
            # TODO: figure out why lines all seem to have a blank line between them, probably something to do with \n at the end or something
            l=OutlineLabel(text=line,size_hint=(None,None),font_size='12sp',halign='left')
            l.bind(texture_size=l.setter('size'))
            layout.add_widget(l)
            sv.scroll_to(l)

    def filesync(self):
        """Callback when the user presses the Sync button. This relies on keyless ssh working, and does most of the work in shell scripts."""
        Logger.info('Config: filesync')
        # set up a popup containing a scrollview to contain stdout output
        layout = GridLayout(cols=1, spacing=1, padding=1, size_hint_y=None)
        layout.bind(minimum_height=layout.setter('height'))
        popup = Popup(title='Sync',size_hint=(0.8,1))
        sv=ScrollView(size_hint=(1,1))
        sv.add_widget(layout)
        popup = Popup(title='Sync',content=sv,size_hint=(0.8,1))
        popup.open()
        # temp file for the rsync, probably a better way to do this
        # this file describes exactly what songs should exist on the host, no more, no less
        tpath=os.path.join(mainconfig.get('paths','tmppath'),"rsync.inc")
        # look in the ini file for all the relevant paths
        synchost = mainconfig.get('sync','synchost')
        syncbasepath = mainconfig.get('sync','syncmusicpath')
        syncfanartpath= mainconfig.get('sync','syncfanartpath')
        basepath = mainconfig.get('paths','musicpath')
        fanartpath= mainconfig.get('paths','fanartpath')
        Logger.info('Filesync: Copying rsync file to carpi')
        # TODO: figure out why this doesn't show up on the screen until after the os.walk has completed
        l=OutlineLabel(text='Copying rsync file to carpi',size_hint=(None,None),font_size='12sp',halign='left')
        l.bind(texture_size=l.setter('size'))
        layout.add_widget(l)
        sv.scroll_to(l)
        # copy the rsync file from the synchost
        # TODO: implement this using python instead of call
        call(["scp",synchost+":rsync.inc",tpath])
        filelist={}
        # use codecs.open to ensure the file is read as utf8, otherwise special chars can be mangled
        with codecs.open(tpath,'r','utf-8') as f:
            for line in f:
                # add each filename to a dict for easy searching later
                filelist[Helpers.decodeFileName(line.rstrip())]=True
        Logger.info('Filesync: Removing old files from carpi')
        # TODO: figure out why this doesn't show up on the screen until after the os.walk has completed
        l=OutlineLabel(text='Removing old files from carpi',size_hint=(None,None),font_size='12sp',halign='left')
        l.bind(texture_size=l.setter('size'))
        layout.add_widget(l)
        sv.scroll_to(l)
        # this whole block walks the filesystem and deletes any file that is not in the rsync file
        # the sync operation is split up into a delete and a copy because that was the only way I could get
        # rsync to work correctly, it was always copying/deleting the wrong things otherwise
        for dirpath, dirnames, filenames in os.walk(Helpers.decodeFileName(basepath)):
            if len(filenames)>0:
                rpath = dirpath[len(basepath+os.sep):]
                for filename in filenames:
                    fpath=os.path.join(Helpers.decodeFileName(rpath),Helpers.decodeFileName(filename))
                    apath = os.path.join(Helpers.decodeFileName(dirpath),Helpers.decodeFileName(filename))
                    if fpath not in filelist:
                        Logger.debug("Filesync: Deleting "+apath)
                        os.remove(Helpers.decodeFileName(apath))
        # TODO: somehow do this all in python instead of shell script, it's ugly
        # also, if the host somehow has tmppath mounted with the no-execute bit set, this will fail
        with open(os.path.join(mainconfig.get('paths','tmppath'),'sync.sh'),'w') as sfile:
            sfile.write("#!/bin/bash\n")
            # delete all empty directories
            sfile.write('find "'+basepath+'" -type d -empty -delete 2>/dev/null\n')
            # copy/update only the files that exist in the rsync file
            sfile.write('rsync -vruxhm --files-from="'+tpath+'" '+synchost+':"'+syncbasepath+'"/ "'+basepath+'"\n')
            # copy the car sticker database to the synchost
            sfile.write('scp /var/lib/mpd/sticker.sql '+synchost+':'+mainconfig.get('sync','synctmppath')+'\n')
            # build a secondary shell script to perform the sticker update on the synchost
            with open(os.path.join(mainconfig.get('paths','tmppath'),'scmd'),'w') as f:
                # attach the copied database and merge sticker data from it to update ratings user has added in the car
                # not using os.path.join here since who knows what os the synchost uses...use linux
                f.write("attach database \""+mainconfig.get('sync','synctmppath')+"/sticker.sql\" as carpi;\n")
                f.write("replace into sticker select * from carpi.sticker;\n")
                f.write("replace into carpi.sticker select * from sticker where name='rating';\n")
                f.write(".quit\n")
            # copy secondary script to synchost and run it
            sfile.write('scp '+os.path.join(mainconfig.get('paths','tmppath'),'scmd')+' '+synchost+':'+mainconfig.get('sync','synctmppath')+'\n')
            # not using os.path.join here since who knows what os the synchost uses...use linux
            sfile.write('ssh -t '+synchost+' sudo sqlite3 /var/lib/mpd/sticker.sql < '+mainconfig.get('sync','synctmppath')+'/scmd\n')
            # copy the now updated sticker database back from the synchost
            # not using os.path.join here since who knows what os the synchost uses...use linux
            sfile.write('scp '+synchost+':'+mainconfig.get('sync','synctmppath')+'/sticker.sql '+mainconfig.get('paths','tmppath')+'\n')
            # build a secondary shell script to perform the sticker update on the host
            with open(os.path.join(mainconfig.get('paths','tmppath'),'scmd'),'w') as f:
                # attach the copied database and merge sticker data from it to update ratings user has added at home
                f.write("attach database \""+os.path.join(mainconfig.get('paths','tmppath'),"sticker.sql")+"\" as carpi;\n")
                f.write("replace into sticker select * from carpi.sticker;\n")
                f.write(".quit\n")
            # run the secondary script
            sfile.write('cat '+os.path.join(mainconfig.get('paths','tmppath'),'scmd')+' | sudo sqlite3 /var/lib/mpd/sticker.sql\n')
            # rsync over all the fanart
            sfile.write('rsync -vruxhm '+synchost+':"'+syncfanartpath+'"/ "'+fanartpath+'"\n')
            # tell mpd about any changes
            sfile.write("mpc update\n")
        # make the shell script executable
        os.chmod(os.path.join(mainconfig.get('paths','tmppath'),'sync.sh'),os.stat(os.path.join(mainconfig.get('paths','tmppath'),'sync.sh')).st_mode|0111)
        # queue for holding stdout
        q = Queue()
        # create a subprocess for the shell script and capture stdout
        p = Popen([os.path.join(mainconfig.get('paths','tmppath'),'sync.sh')],stdout=PIPE,bufsize=1,close_fds=True)
        # check the stdout queue every .1 seconds and write lines to the scrollview if there is output
        event=Clock.schedule_interval(partial(self.write_queue_line,q,layout,sv),0.1)
        # run all this crap in a separate thread to be non-blocking
        t = Thread(target=self.enqueue_output,args=(p.stdout,q,event,popup,tpath,synchost,layout,sv))
        t.daemon = True
        t.start()



import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.logger import Logger
from twisted.internet.defer import inlineCallbacks
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.properties import BooleanProperty
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.clock import Clock
from copy import deepcopy
from functools import partial
import os

from kmpc.extra import KmpcHelpers, OutlineTabbedPanelItem
import kmpc.kmpcapp

Helpers=KmpcHelpers()

class LibraryTabbedPanelItem(OutlineTabbedPanelItem):
    """The Library tab, for browsing through mpd's library."""

    # set some initial variables
    current_view = {'value': 'root', 'base':'/','info':{'type':'uri'}}
    library_selection = {}

    def __init__(self,**kwargs):
        super(self.__class__,self).__init__(**kwargs)
        self.app=App.get_running_app().root

    def change_view_type(self,value):
        """Callback when user presses one of the Library view buttons."""
        Logger.info("Library: View changed to "+value)
        # make sure nothing is selected in the recyclebox
        self.rbl.clear_selection()
        # the button state things sort of implement a tabbed view without it being a tabbedview
        # this is so there doesn't have to be a separate recyclebox for each view type
        # probably a more elegant way of doing this
        if value == 'Files':
            self.current_view = {'value': 'root','base':'/','info':{'type':'uri'}}
            kmpc.kmpcapp.mainmpdconnection.protocol.lsinfo(self.current_view['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='down'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='normal'
            self.ids.delete_button.disabled=True
        elif value == 'Albums':
            self.current_view = {'value': 'All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'}}
            kmpc.kmpcapp.mainmpdconnection.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='down'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='normal'
            self.ids.delete_button.disabled=True
        elif value == 'Tracks':
            self.current_view = {'value': 'All Track Artists','base':'All Track Artists','info':{'type':'roottracks'}}
            kmpc.kmpcapp.mainmpdconnection.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='down'
            self.ids.playlists_button.state='normal'
            self.ids.delete_button.disabled=True
        elif value == 'Playlists':
            self.current_view = {'value':'All Playlists','base':'All Playlists','info':{'type':'playlist'}}
            kmpc.kmpcapp.mainmpdconnection.protocol.listplaylists().addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='down'
            self.ids.delete_button.disabled=False

    def reload_view(self,result):
        """Callback that loads library data from mpd into the recyclebox."""
        Logger.info("Library: reload_view() current type: "+self.current_view['info']['type'])
        # clear all current recycleview data
        self.rv.data=[]
        if self.current_view['info']['type'] == 'uri':
            # file-based browsing uses uri types
            if self.current_view['base'] != '/':
                # if we aren't at the base of the filesystem, allow browsing to parent folder
                (hbase,tbase)=os.path.split(self.current_view['base'])
                (b1,b2)=os.path.split(hbase)
                if b2 == '':
                    b2 = 'root'
                upbase=os.path.normpath(self.current_view['base']+'/..')
                if upbase == '.':
                    upbase = '/'
                r = {'value':"up to "+b2,'base':upbase,'info':{'type':'uri'}}
                self.rv.data.append(r)
                self.current_header.text = tbase
            else:
                # at filesystem base
                self.current_header.text = 'All Files'
        elif self.current_view['info']['type'] == 'albumartistsort':
            # inside a particular album artist, allow browsing to root
            r={'value': 'up to All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'}}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'album':
            # inside a particular album, allow browsing to album artist
            r={'value':'up to '+self.current_view['info']['albumartistsort'],'base':self.current_view['info']['albumartistsort'],'info':{'type':'albumartistsort'}}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'artistsort':
            # inside a particular artist, allow browsing to root
            r={'value': 'up to All Track Artists','base':'All Track Artists','info':{'type':'roottracks'}}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        else:
            # something else?
            self.current_header.text = self.current_view['base']
        # loop through the rows mpd returned and add them to the recycleview
        for row in result:
            if 'playlist' in row:
                # we found a playlist, display it
                if self.current_view['info']['type'] != 'uri':
                    Logger.debug("Library: playlist found = "+row['playlist'])
                    r = {'value':row['playlist'],'base':row['playlist'],'info':{'type':'playlist'}}
                    self.rv.data.append(r)
            elif 'directory' in row:
                # we found a directory, format and display it
                Logger.debug("Library: directory found: ["+row['directory']+"]")
                (b1,b2)=os.path.split(row['directory'])
                r={'value':b2,'base':row['directory'],'info':{'type':'uri'}}
                self.rv.data.append(r)
            elif 'file' in row:
                # we found a song, format and display it
                Logger.debug("FileBrowser: file found: ["+row['file']+"]")
                r={'value':Helpers.formatsong(row),'base':row['file'],'info':{'type':'file'}}
                self.rv.data.append(r)
            else:
                # we found something else, figure out what it is based on our current view type
                if self.current_view['info']['type'] == 'rootalbums':
                    # we found an album artist, display it
                    Logger.debug("Library: album artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'albumartistsort'}}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'albumartistsort':
                    # we found an album, display it
                    Logger.debug("Library: album found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'album','albumartistsort':self.current_view['base']}}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'roottracks':
                    # we found an artist, display it
                    Logger.debug("Library: track artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'artistsort'}}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'artistsort':
                    # we found a track, display it
                    # note that tracks are not parsed by Helpers.formatsong like filenames are
                    # mpd formats a track itself
                    Logger.debug("Library: track found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'track','artistsort':self.current_view['base']}}
                    self.rv.data.append(r)
                else:
                    # shouldn't ever see this, but y'know, whatever
                    Logger.warn("Library: not sure what to do with ["+format(row)+"]")

    def handle_long_touch(self,row,index):
        """Callback for handling long touches in the recycleview."""
        Logger.debug("Library: handle_long_touch("+format(row)+")")
        # load up the selected row as the current view, then display it
        self.current_view = deepcopy(row)
        if row['info']['type'] == 'uri':
            # selected a directory
            kmpc.kmpcapp.mainmpdconnection.protocol.lsinfo(row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'rootalbums':
            # selected an album artist
            kmpc.kmpcapp.mainmpdconnection.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'albumartistsort':
            # selected an album
            kmpc.kmpcapp.mainmpdconnection.protocol.list('album','albumartistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'album':
            # selected an album track
            kmpc.kmpcapp.mainmpdconnection.protocol.find('album',row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'roottracks':
            # selected an artist
            kmpc.kmpcapp.mainmpdconnection.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'artistsort':
            # selected a track
            kmpc.kmpcapp.mainmpdconnection.protocol.list('title','artistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'playlist':
            # selected a playlist, go ahead and load it up and play it
            kmpc.kmpcapp.mainmpdconnection.protocol.clear()
            kmpc.kmpcapp.mainmpdconnection.protocol.load(row['base'])
            kmpc.kmpcapp.mainmpdconnection.protocol.play('0')
        elif row['info']['type'] == 'file':
            # selected a file, append it to the playlist and play from there
            kmpc.kmpcapp.mainmpdconnection.protocol.clear()
            (a,b)=os.path.split(row['base'])
            kmpc.kmpcapp.mainmpdconnection.protocol.add(a)
            kmpc.kmpcapp.mainmpdconnection.protocol.play(str(int(index)-1))
        else:
            # should never see this
            Logger.warn("Library: long-touch for ["+format(row)+"] not implemented")

    def handle_mpd_error(self,result):
        """Callback for handling mpd exceptions."""
        Logger.error('Library: MPDIdleHandler Callback error: {}'.format(result))

    def browser_add_find(self,result):
        """Callback for appending a bunch of tracks to the playlist."""
        for rrow in result:
            kmpc.kmpcapp.mainmpdconnection.protocol.add(rrow['file'])

    def browser_add_find_one(self,result):
        """Callback for appending one track to the playlist."""
        # since mpd always returns a list, just do the first one then break
        for rrow in result:
            kmpc.kmpcapp.mainmpdconnection.protocol.add(rrow['file'])
            break

    def browser_add(self,clearfirst,insert):
        """Callback when user presses the '+' (add), '!' (clear then add), or '>' (insert) buttons."""
        Logger.info('Library: browser_add('+str(clearfirst)+')')
        # if !, clear the playlist
        if clearfirst:
            Logger.info('Library: Clearing playlist')
            kmpc.kmpcapp.mainmpdconnection.protocol.clear()
        # loop through the recyclebox and add each selected node
        for index in self.rbl.selected_nodes:
            row = self.rv.data[index]
            mtype=row['info']['type']
            Logger.info("Library: Adding "+mtype+" '"+row['base']+"' to current playlist")
            if mtype == 'uri' or mtype == 'file':
                if insert and self.app.currsong:
                    # if >, insert the song/directory after the currently playing song
                    kmpc.kmpcapp.mainmpdconnection.protocol.addid(row['base'],str(int(self.app.currsong)+1))
                else:
                    # append the song/directory to the playlist
                    kmpc.kmpcapp.mainmpdconnection.protocol.add(row['base'])
            elif mtype == 'albumartistsort':
                # append all tracks by a particular album artist
                kmpc.kmpcapp.mainmpdconnection.protocol.find(mtype,row['base']).addCallback(self.browser_add_find).addErrback(self.handle_mpd_error)
            elif mtype == 'album':
                # append all tracks on a particular album
                kmpc.kmpcapp.mainmpdconnection.protocol.find(mtype,row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(self.browser_add_find).addErrback(self.handle_mpd_error)
            elif mtype == 'artistsort':
                # append all tracks by a particular artist
                kmpc.kmpcapp.mainmpdconnection.protocol.find(mtype,row['base']).addCallback(self.browser_add_find).addErrback(self.handle_mpd_error)
            elif mtype == 'track':
                # append a particular artist's specific track
                # currently just adds the first match that mpd finds
                kmpc.kmpcapp.mainmpdconnection.protocol.find('artistsort',row['info']['artistsort'],'title',row['base']).addCallback(self.browser_add_find_one).addErrback(self.handle_mpd_error)
            elif mtype == 'playlist':
                # append a playlist
                kmpc.kmpcapp.mainmpdconnection.protocol.load(row['base'])
            else:
                # should never see this
                Logger.warning("Library: "+mtype+' not implemented')
        # clear the currently selected rows after doing the above work
        self.rbl.clear_selection()

    def browser_delete(self):
        """Callback when user presses the delete playlist button."""
        for index in self.rbl.selected_nodes:
            plname=self.rv.data[index]['base']
            Logger.info("Library: deleting playlist "+plname)
            kmpc.kmpcapp.mainmpdconnection.protocol.rm(plname).addErrback(self.handle_mpd_error)
            self.current_view = {'value':'All Playlists','base':'All Playlists','info':{'type':'playlist'}}
            kmpc.kmpcapp.mainmpdconnection.protocol.listplaylists().addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        self.rbl.clear_selection()

class LibraryRecycleBoxLayout(LayoutSelectionBehavior,RecycleBoxLayout):
    """Adds selection and focus behaviour to a recyclebox."""

class LibraryRow(RecycleDataViewBehavior,BoxLayout):
    """Adds selection support to a row in a recycleview."""
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def __init__(self,**kwargs):
        super(self.__class__,self).__init__(**kwargs)
        self.app=App.get_running_app().root

    def long_touch(self, touch, index, *args):
        """Callback when user long-presses on a row."""
        Logger.debug("Library: long-touch on "+str(index))
        self.app.ids.library_tab.rbl.clear_selection()
        self.app.ids.library_tab.handle_long_touch(self.app.ids.library_tab.rv.data[index],index)

    def refresh_view_attrs(self, rv, index, data):
        """Catch and handle the view changes."""
        self.index = index
        return super(self.__class__, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        """Adds selection, long-press handling on touch down."""
        if super(self.__class__, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            # these lines start a 1 second clock to detect long-presses
            callback = partial(self.long_touch, touch, self.index)
            Clock.schedule_once(callback, 1)
            touch.ud['event'] = callback
            return self.parent.select_with_touch(self.index, touch)

    def on_touch_up(self, touch):
        """Clean up long-press handling on touch up."""
        if super(self.__class__, self).on_touch_up(touch):
            return True
        # if i don't check for this, the app crashes when things scroll off screen
        if 'event' in touch.ud:
            Clock.unschedule(touch.ud['event'])

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected
        lt=self.app.ids.library_tab
        if is_selected:
            lt.library_selection[index] = True
        else:
            if index in lt.library_selection:
                del lt.library_selection[index]



# import dependencies
import os
import traceback
import mutagen
import io
import random
import socket
from pkg_resources import resource_filename
from functools import partial

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

# import all the other kivy stuff
from kivy.config import Config
from kivy.app import App
from kivy.logger import Logger
from kivy.graphics import Color,Rectangle
from kivy.core.image import Image as CoreImage
from kivy.metrics import Metrics, sp
from kivy.clock import Clock
from kivy.uix.widget import Widget
from kivy.uix.tabbedpanel import TabbedPanel, TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.scrollview import ScrollView
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import Image,AsyncImage
from kivy.uix.popup import Popup
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.slider import Slider
from kivy.factory import Factory
from kivy.lang import Builder
from kivy.properties import ObjectProperty

# import our local modules
from kmpc.extra import KmpcHelpers,ExtraSlider,ClearButton,OutlineLabel,OutlineButton,MpdConnection
from kmpc.playlistpanel import PlaylistTabbedPanelItem

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

# load the interface.kv file
Builder.load_file(resource_filename(__name__,os.path.join('resources','interface.kv')))

Helpers=KmpcHelpers()

class KmpcInterface(TabbedPanel):
    """The main class that ties it all together."""

    def __init__(self,**kwargs):
        """Zero out variables, pull in config file, connect to mpd."""
        super(self.__class__,self).__init__(**kwargs)
        # bind callbacks for tab changes
        self.bind(current_tab=self.main_tab_changed)
        self.mpd_status={'state':'stop','repeat':0,'single':0,'random':0,'consume':0,'curpos':0}
        self.currsong=None
        self.nextsong=None
        self.currfile=None
        self.track_slider_task=None
        self.accessoryPopup=Factory.AccessoryPopup()
        self.tcolor=1
        self.ocolor=0
        self.settingsPopup=Factory.SettingsPopup()
        global mainmpdconnection
        global mainconfig
        mainconfig=Helpers.loadconfigfile()
        mainmpdconnection=MpdConnection(mainconfig,self.mpd_idle_handler,[self.init_mpd])

    def settings_popup(self):
        self.settingsPopup.open()
        # get the host's IP address and display it
        self.settingsPopup.ids.ip_label.text="IP Address: "+format(self.get_ip())
        mainmpdconnection.protocol.status().addCallback(partial(self.update_mixers,self.settingsPopup)).addErrback(self.handle_mpd_error)

    def update_mixers(self,p,result):
        # set up the crossfade slider
        if 'xfade' in result:
            v = int(result['xfade'])
        else:
            v = 0
        p.ids.crossfade_slider.value=v
        # set up the mixrampdb slider
        if 'mixrampdb' in result:
            v = round(float(result['mixrampdb']),6)
        else:
            v = 0.0
        p.ids.mixrampdb_slider.value=float(str(v)[1:])
        # set up the mixrampdelay slider
        if 'mixrampdelay' in result:
            v = round(float(result['mixrampdelay']),6)
        else:
            v = 0.0
        p.ids.mixrampdelay_slider.value=v

    def change_text_color(self,color):
        Logger.debug("NowPlaying: change_text_color to "+format(color))
        self.tcolor=color
        def _tc(widget):
            for child in widget.children:
                _tc(child)
            if issubclass(widget.__class__,OutlineLabel):
                widget.color=[color,color,color,1]
        for child in self.children:
            _tc(child)

    def change_outline_color(self,color):
        Logger.debug("NowPlaying: change_outline_color to "+format(color))
        self.ocolor=color
        def _tc(widget):
            for child in widget.children:
                _tc(child)
            if issubclass(widget.__class__,OutlineLabel):
                widget.outline_color=[color,color,color,1]
        for child in self.children:
            _tc(child)

    def change_crossfade(self,v):
        """Callback when user changes crossfade slider."""
        Logger.info('Settings: change_crossfade')
        mainmpdconnection.protocol.crossfade(str(v)).addErrback(self.handle_mpd_error)

    def change_mixrampdb(self,v):
        """Callback when user changes mixrampdb slider."""
        Logger.info('Settings: change_mixrampdb')
        mainmpdconnection.protocol.mixrampdb(str(0.0-v)).addErrback(self.handle_mpd_error)

    def change_mixrampdelay(self,v):
        """Callback when user changes mixrampdelay slider."""
        Logger.info('Settings: change_mixrampdelay')
        mainmpdconnection.protocol.mixrampdelay(str(v)).addErrback(self.handle_mpd_error)

    def get_ip(self):
        """Method that tries to get the local IP address, and returns localhost if there isn't one."""
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

    def main_tab_changed(self,obj,value):
        """Callback when top tab is changed."""
        self.active_tab = value.text
        Logger.info("Application: Changed active tab to "+self.active_tab)
        # Playlist is the only one that needs help when activating
        if self.active_tab == 'Playlist':
            # switching to the playlist tab repopulates it if it is empty
            if len(self.ids.playlist_tab.rv.data) == 0:
                mainmpdconnection.protocol.playlistinfo().addCallback(self.ids.playlist_tab.populate_playlist).addErrback(self.ids.playlist_tab.handle_mpd_error)

    def handle_mpd_error(self,result):
        mainmpdconnection.handle_mpd_error(result)

    def handle_mpd_message(self,result):
        """Callback for mpd 'kmpc' messages."""
        # result is an iterator although it only generally has one member
        for m in result:
            message = m['message']
            Logger.info("Application: MPD Message: "+message)
            if message == 'AccessoryOff':
                # generated by mausberry circuit, accessory power is off
                self.accessoryPopup.open()
            elif message == 'AccessoryOn':
                # generated by mausberry circuit, accessory power is on
                self.accessoryPopup.dismiss()

    def current_track_slider_release(self):
        """Callback when track slider is released."""
        # get the value of the track slider
        curpos = int(self.ids.current_track_slider.value)
        Logger.debug("Application: touch up on track slider at "+str(curpos))
        # seek to that position in the track
        mainmpdconnection.protocol.seekcur(str(curpos)).addErrback(self.handle_mpd_error)

    def current_track_slider_down(self):
        """Callback when track slider is pressed."""
        Logger.debug("Application: touch down on track slider")
        # cancel the once-per-second update while finger is down so it doesn't move out from under it
        self.track_slider_task.cancel()

    def update_track_slider(self,dt):
        """Increment the slider by 1 second every second."""
        Logger.debug("Application: update_track_slider")
        curpos=int(self.ids.current_track_slider.value)+1
        self.ids.current_track_slider.value=curpos

    def stop_zero_stuff(self):
        """Clear the screen and variables if playback is fully stopped."""
        self.ids.current_track_slider.value=0
        self.ids.current_track_slider.max=0
        self.ids.current_playlist_track_number_label.text=''
        self.ids.next_song_artist_label.text = ''
        self.currfile = None
        self.currsong = None
        self.nextsong = None
        self.ids.song_star_layout.clear_widgets()
        self.ids.album_cover_layout.clear_widgets()
        self.ids.trackinfo.clear_widgets()
        lbl = OutlineLabel(text="Playback Stopped")
        self.ids.trackinfo.add_widget(lbl)
        self.ids.player.canvas.before.add(Rectangle(source=resource_filename(__name__,os.path.join("resources","backdrop.png")),size=self.ids.player.size,pos=self.ids.player.pos))

    def update_mpd_status(self,result):
        """Callback when mpd status changes."""
        Logger.debug('NowPlaying: update_mpd_status()')
        # probably state is the only one necessary, but hey
        self.mpd_status['state']=result['state']
        self.mpd_status['repeat']=result['repeat']
        self.mpd_status['single']=result['single']
        self.mpd_status['random']=result['random']
        self.mpd_status['consume']=result['consume']
        if self.mpd_status['state'] == 'stop':
            # if stopped, there are no current or next songs
            self.currsong=None
            self.nextsong=None
        else:
            # save current song, this is a 0-based index into the playlist
            self.currsong=result['song']
            # ask mpd for current song data
            mainmpdconnection.protocol.currentsong().addCallback(self.update_mpd_currentsong).addErrback(self.handle_mpd_error)
            # save next song, this is a 0-based index into the playlist
            if 'nextsong' in result:
                self.nextsong=result['nextsong']
                # ask mpd for next song data
                mainmpdconnection.protocol.playlistinfo(self.nextsong).addCallback(self.update_mpd_nextsong).addErrback(self.handle_mpd_error)
            else:
                self.nextsong=None
        stflags=['normal','down']
        # check various flag states
        self.ids.repeat_button.state=stflags[int(self.mpd_status['repeat'])]
        self.ids.single_button.state=stflags[int(self.mpd_status['single'])]
        self.ids.random_button.state=stflags[int(self.mpd_status['random'])]
        self.ids.consume_button.state=stflags[int(self.mpd_status['consume'])]
        if self.mpd_status['state']=='pause' or self.mpd_status['state']=='stop':
            # play/pause button should be a play button that is unpressed
            self.ids.play_button.state='normal'
            self.ids.play_button.text=u"\uf04b"
            self.track_slider_task.cancel()
        else:
            # play/pause button should be a pause button that is pressed
            self.ids.play_button.state='down'
            self.ids.play_button.text=u"\uf04c"
        if self.mpd_status['state'] == 'stop':
            # zero everything out if we are stopped
            self.stop_zero_stuff()
        else:
            # mpd returns {elapsed seconds}:{total seconds}, the following splits each to minute:second
            c,t=result['time'].split(":")
            # set the max slider value to the total seconds
            self.ids.current_track_slider.max = int(t)
            # set the current slider value to the current seconds
            self.mpd_status['curpos']=int(c)
            self.ids.current_track_slider.value = int(c)
            if self.mpd_status['state']!='pause':
                self.track_slider_task()
            # throws an exception if i don't do this
            a=int(result['song'])+1
            b=int(result['playlistlength'])
            self.ids.current_playlist_track_number_label.text = "%d of %d" % (a,b)
        # update the playlist tab with status results
        self.ids.playlist_tab.update_mpd_status(result)

    def change_artist_image(self,img,al_path,instance):
        """Called when you click on an artist logo, changes it to another at random."""
        Logger.debug("NowPlaying: change_artist_image (current path is "+img.source+")")
        img_path=img.source
        while img_path==img.source:
            img_path=os.path.join(al_path,random.choice(os.listdir(al_path)))
        if os.path.isfile(img_path):
            img.source=img_path
        Logger.debug("NowPlaying: change_artist_image (new path is "+img.source+")")

    def update_mpd_currentsong(self,result):
        """Callback for mpd currentsong data."""
        Logger.debug('NowPlaying: update_mpd_currentsong()')
        # this is so expensive screen updates only happen if the song has changed since the last time this callback was called
        songchange=False
        # ti is the track info widget
        ti=self.ids.trackinfo
        # if result is undefined, there's not actually a song playing
        if result:
            # if class's current file doesn't match what mpd returns, the song has changed
            if self.currfile != result['file']:
                songchange=True
            # update class's current file
            self.currfile = result['file']
            if songchange:
                # clear the track info widget
                ti.clear_widgets()
                # clear the album cover
                self.ids.album_cover_layout.clear_widgets()
                # get the stored star rating
                mainmpdconnection.protocol.sticker_get('song',self.currfile,'rating').addCallback(self.update_mpd_sticker_rating).addErrback(self.handle_mpd_no_sticker)
                # figure out the full path of the file
                bp=mainconfig.get('paths','musicpath')
                # p is the absolute path
                p=os.path.join(bp,result['file'])
                haslogo=False
                # set the release year if mpd has it
                year=None
                if 'date' in result:
                    year=result['date'][:4]
                # pull the fanart folder from ini file
                fa_path=mainconfig.get('paths','fanartpath')
                # try to get the artistid, set it to 0000 if it doesn't exist
                try:
                    aids=str(result['musicbrainz_artistid'])
                except KeyError:
                    aids='0000'
                # if a track has multiple artists, split it up
                artistids=aids.split('/')
                # create a widget to hold the cover art
                cbl = BoxLayout(size_hint=(1,1),orientation='horizontal')
                # loop through artistids
                for mb_aid in artistids:
                    # try to find a logo for this artist in the fanart folder
                    try:
                        # look in the 'logo' subfolder of the artistid path
                        al_path=os.path.join(fa_path,mb_aid,'logo')
                        # pick one at random
                        img_path=os.path.join(al_path,random.choice(os.listdir(al_path)))
                        if os.path.isfile(img_path):
                            # create an image button out of the logo so you can press it
                            img = ImageButton(source=os.path.join(al_path,img_path),allow_stretch=True,color=(1,1,1,0.65))
                            img.bind(on_press=partial(self.change_artist_image,img,al_path))
                            cbl.add_widget(img)
                            haslogo=True
                    except Exception as e:
                        Logger.exception("update_mpd_currentsong: "+format(e))
                if haslogo:
                    # we found a logo, show it
                    current_artist_label = cbl
                else:
                    # no logo found, just print the artist name instead
                    current_artist_label = InfoLargeLabel(text = result['artist'],font_size=Helpers.getfontsize(result['artist']))
                # clear the background of the player canvas so we can show the artist background
                self.ids.player.canvas.before.clear()
                # pick an artistid at random for the background
                mb_aid=random.choice(artistids)
                try:
                    # look in the 'artistbackground' subfolder of the artistid path
                    ab_path=os.path.join(fa_path,mb_aid,'artistbackground')
                    # pick one at random
                    img_path=random.choice(os.listdir(ab_path))
                    # update the player background with the image
                    self.ids.player.canvas.before.add(Rectangle(source=os.path.join(ab_path,img_path),size=self.ids.player.size,pos=self.ids.player.pos))
                except:
                    # update the player background with the default backdrop
                    self.ids.player.canvas.before.add(Rectangle(source=resource_filename(__name__,os.path.join("resources","backdrop.png")),size=self.ids.player.size,pos=self.ids.player.pos))
                if os.path.isfile(p):
                    Logger.debug('NowPlaying: found good file at path '+p)
                    # load up the file to read the tags
                    f = mutagen.File(p)
                    cimg = None
                    data = None
                    originalyear = None
                    # if config file says use originalyear, use it instead of mpd's year
                    if mainconfig.getboolean('flags','originalyear'):
                        if 'TXXX:originalyear' in f.keys():
                            originalyear=format(f['TXXX:originalyear'])
                    # try to get mp3 cover, if this throws an exception it's not an mp3 or it doesn't have a cover
                    try:
                        pframes = f.tags.getall("APIC")
                        # id3v2 can store any number of image frames, we just want the first one
                        for frame in pframes:
                            ext = 'img'
                            # figure out the file type
                            if frame.mime.endswith('jpeg') or frame.mime.endswith('jpg'):
                                ext = 'jpg'
                            elif frame.mime.endswith('png'):
                                ext = 'png'
                            elif frame.mime.endswith('bmp'):
                                ext = 'bmp'
                            elif frame.mime.endswith('gif'):
                                ext = 'gif'
                            # pull the raw image data into a variable
                            data=io.BytesIO(bytearray(frame.data))
                            break
                    except AttributeError:
                        pass
                    # try to get mp4 cover
                    if 'covr' in f.keys():
                        # figure out the file type
                        if f['covr'][0].imageformat == mutagen.mp4.MP4Cover.FORMAT_PNG:
                            ext = 'png'
                        else:
                            ext = 'jpg'
                        # pull the raw image data into a variable
                        data=io.BytesIO(bytearray(f['covr'][0]))
                    if data:
                        # if we got image data, load it as a kivy.core.image
                        cimg = CoreImage(data,ext=ext)
                    if mainconfig.getboolean('flags','originalyear') and originalyear and year and int(originalyear)!=int(year):
                        tt="["+originalyear+"]\n{"+year+"}"
                    elif year:
                        tt="["+year+"]"
                    else:
                        tt=""
                    if cimg:
                        # TODO: sometimes this just returns a black rectangle, i think i need to catch more specific exceptions
                        # and figure out what exactly is happening
                        img=CoverButton(img=cimg,layout=self.ids.album_cover_layout,text=tt,halign='center')
                    else:
                        img=CoverButton(img=CoreImage(resource_filename(__name__,os.path.join('resources','clear.png'))),layout=self.ids.album_cover_layout,text=tt,halign='center')
                    self.ids.album_cover_layout.add_widget(img)
                    # popup the cover large if you press it
                    img.bind(on_press=self.cover_popup)
                else:
                    # this should _probably_ never happen
                    Logger.debug('NowPlaying: no file found at path '+p)
                # add the correct artist name widget
                ti.add_widget(current_artist_label)
                if haslogo:
                    # we found an artist logo, put the song and album labels in a separate boxlayout to separate them a bit
                    lyt = BoxLayout(orientation='vertical',padding_y='10sp')
                    current_song_label = InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title']))
                    lyt.add_widget(current_song_label)
                    current_album_label = InfoLargeLabel(text = result['album'],font_size=Helpers.getfontsize(result['album']))
                    lyt.add_widget(current_album_label)
                    ti.add_widget(lyt)
                else:
                    # no artist logo, just add the song and album labels directly to the track info widget
                    current_song_label = InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title']))
                    ti.add_widget(current_song_label)
                    current_album_label = InfoLargeLabel(text = result['album'],font_size=Helpers.getfontsize(result['album']))
                    ti.add_widget(current_album_label)
        else:
            # there's not a current song, so zero everything out
            self.stop_zero_stuff()

    def update_mpd_sticker_rating(self,result):
        """Callback for song that has a rating in mpd."""
        Logger.debug('NowPlaying: update_mpd_sticker_rating')
        # make a clear button for the star rating
        btn = ClearButton(padding_x='10sp',font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')),halign='center',valign='middle',markup=True)
        # look up the correct string for the rating
        btn.text = Helpers.songratings(mainconfig)[result]['stars']
        # bind the popup for setting rating
        btn.bind(on_press=self.rating_popup)
        # clear the layout widget and add the new one
        self.ids.song_star_layout.clear_widgets()
        self.ids.song_star_layout.add_widget(btn)

    def handle_mpd_no_sticker(self,result):
        """Callback for song that has no rating in mpd."""
        Logger.debug('NowPlaying: handle_mpd_no_sticker')
        # make a clear button for the star rating
        btn = ClearButton(padding_x='10sp',font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')),halign='center',valign='middle',markup=True)
        # set the string to the circled question mark icon
        btn.text = u"\uf29c"
        # bind the popup for setting rating
        btn.bind(on_press=self.rating_popup)
        # clear the layout widget and add the new one
        self.ids.song_star_layout.clear_widgets()
        self.ids.song_star_layout.add_widget(btn)

    def update_mpd_nextsong(self,result):
        """Callback for next song data from mpd."""
        Logger.debug('NowPlaying: update_mpd_nextsong()')
        # result is a list with one member
        for obj in result:
            # set the next song label
            self.ids.next_song_artist_label.text = 'Up Next: '+obj['artist']+' - '+obj['title']

    def prev_pressed(self):
        """Callback for prev button pressed."""
        Logger.debug('Application: prev_pressed()')
        mainmpdconnection.protocol.previous()

    def play_pressed(self):
        """Callback for play/pause button pressed."""
        Logger.debug('Application: play_pressed()')
        if self.mpd_status['state'] == 'play':
            # pause if playing
            mainmpdconnection.protocol.pause()
        else:
            # play if paused or stopped
            mainmpdconnection.protocol.play()

    def next_pressed(self):
        """Callback for next button pressed."""
        Logger.debug('Application: next_pressed()')
        mainmpdconnection.protocol.next()

    def repeat_pressed(self):
        """Callback for repeat button pressed."""
        Logger.debug('Application: repeat_pressed()')
        # toggle on/off
        mainmpdconnection.protocol.repeat(str(1-int(self.mpd_status['repeat'])))

    def single_pressed(self):
        """Callback for single button pressed."""
        Logger.debug('Application: single_pressed()')
        # toggle on/off
        mainmpdconnection.protocol.single(str(1-int(self.mpd_status['single'])))

    def random_pressed(self):
        """Callback for random button pressed."""
        Logger.debug('Application: random_pressed()')
        # toggle on/off
        mainmpdconnection.protocol.random(str(1-int(self.mpd_status['random'])))

    def consume_pressed(self):
        """Callback for consume button pressed."""
        Logger.debug('Application: consume_pressed()')
        # toggle on/off
        mainmpdconnection.protocol.consume(str(1-int(self.mpd_status['consume'])))

    def rating_popup(self,instance):
        """Popup for setting song rating."""
        Logger.debug('Application: rating_popup()')
        # create a layout and add it to the popup
        layout = GridLayout(cols=2,spacing=10)
        popup = Popup(title='Rating',content=layout,size_hint=(0.8,1))
        # loop from 0-10
        for r in list(range(0,11)):
            # make a button
            btn=Button(font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')))
            # look up the correct string for the rating
            btn.text=Helpers.songratings(mainconfig)[str(r)]['stars']
            # set some widget variables
            btn.rating=str(r)
            btn.popup=popup
            # add the button to the layout
            layout.add_widget(btn)
            # bind the button press to set the rating
            btn.bind(on_press=self.rating_set)
            # add a label to explain the ratings, this is pretty subjective
            lbl=OutlineLabel(text=Helpers.songratings(mainconfig)[str(r)]['meaning'],halign='left')
            # add the label to the layout
            layout.add_widget(lbl)
        # pop it on up, if you press outside the popup it just goes away without setting the rating
        popup.open()

    def rating_set(self,instance):
        """Method that sets a song's rating."""
        Logger.debug('Application: rating_set('+instance.rating+')')
        # close the rating popup
        instance.popup.dismiss()
        # tell mpd to set the rating sticker
        mainmpdconnection.protocol.sticker_set('song',self.currfile,'rating',instance.rating)

    def cover_popup(self,instance):
        """Popup for showing a larger version of the album cover."""
        Logger.debug('Application: cover_popup()')
        layout = BoxLayout()
        popup = Popup(title='Cover',content=layout,size_hint=(0.6,1))
        # pull the already loaded image texture
        img = Image(texture=instance.img.texture,allow_stretch=True)
        # add it to the layout
        layout.add_widget(img)
        # pop pop pop
        popup.open()

    def change_backlight(self,value):
        """Method that sets the backlight to a certain value."""
        Logger.info('Application: change_backlight('+str(value)+') rpienable = '+mainconfig.get('flags','rpienable'))
        # only if the ini file says it's ok
        if mainconfig.getboolean('flags','rpienable'):
            import rpi_backlight as bl
            # set the brightness
            bl.set_brightness(int(value), smooth=True, duration=1)

    def init_mpd(self,instance):
        # get the initial mpd status
        mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(mainmpdconnection.handle_mpd_error)
        # create the once-per-second update of the track slider
        self.track_slider_task=Clock.schedule_interval(self.update_track_slider,1)
        self.track_slider_task.cancel()
        # subscribe to 'kmpc' to check for messages from mpd
        mainmpdconnection.protocol.subscribe('kmpc')

    def mpd_idle_handler(self,result):
        # notify various subsystems based on what changed
        for s in result:
            Logger.info('MPDIdleHandler: Changed '+format(s))
            if format(s) == 'playlist':
                # playlist was changed, ask mpd for playlist info
                mainmpdconnection.protocol.playlistinfo().addCallback(self.ids.playlist_tab.populate_playlist).addErrback(self.ids.playlist_tab.handle_mpd_error)
                # force a reload of nextsong if playlist changes
                self.nextsong = None
                mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
            elif format(s) == 'player':
                # player was changed, ask mpd for player status
                mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
            elif format(s) == 'sticker':
                # song rating sticker was changed, ask mpd for current song rating
                mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
                mainmpdconnection.protocol.sticker_get('song',self.currfile,'rating').addCallback(self.update_mpd_sticker_rating).addErrback(self.handle_mpd_no_sticker)
            elif format(s) == 'options':
                # some playback option was changed, ask mpd for player status
                mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
            elif format(s) == 'message':
                # an mpd message was received, ask mpd what it was
                mainmpdconnection.protocol.readmessages().addCallback(self.handle_mpd_message).addErrback(self.handle_mpd_error)
            else:
                # default if none of the above, ask mpd for player status
                mainmpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)

class KmpcApp(App):
    """The overall app class, builds the main interface widget."""

    def __init__(self,args):
        """Override kivy config values with necessary ones"""
        Config.set('kivy','keyboard_mode','systemanddock')
        Config.set('graphics','width',800)
        Config.set('graphics','height',480)
        super(self.__class__,self).__init__()

    def build(self):
        """Instantiates KmpcInterface."""
        # setup some variables that interface.kv will use
        # this is necessary to support packaging the app
        self.normalfont = resource_filename(__name__,os.path.join('resources','DejaVuSans.ttf'))
        self.boldfont = resource_filename(__name__,os.path.join('resources','DejaVuSans-Bold.ttf'))
        self.fontawesomefont = resource_filename(__name__,os.path.join('resources','FontAwesome.ttf'))
        self.buttonnormal = resource_filename(__name__,os.path.join('resources','button-normal.png'))
        self.buttondown = resource_filename(__name__,os.path.join('resources','button-down.png'))
        self.clear = resource_filename(__name__,os.path.join('resources','clear.png'))
        self.backdrop = resource_filename(__name__,os.path.join('resources','backdrop.png'))
        self.listbackdrop = resource_filename(__name__,os.path.join('resources','list-backdrop.png'))
        self.listbackdropselected = resource_filename(__name__,os.path.join('resources','list-backdrop-selected.png'))
        self.trackslidercursor = resource_filename(__name__,os.path.join('resources','track-slider-cursor.png'))
        return KmpcInterface()

class InfoLargeLabel(OutlineLabel):
    """A label with large text."""
    pass

class InfoSmallLabel(OutlineLabel):
    """A label with small text."""
    pass

class ImageButton(ButtonBehavior, AsyncImage):
    """An image that you can press."""
    pass

class CoverButton(OutlineButton):
    img = ObjectProperty(None)
    layout = ObjectProperty(None)

    def __init__(self,**kwargs):
        super(self.__class__,self).__init__(**kwargs)
        self.background_normal = resource_filename(__name__,os.path.join('resources','clear.png'))
        self.background_down = resource_filename(__name__,os.path.join('resources','clear.png'))
        self.font_name = resource_filename(__name__,os.path.join('resources','DejaVuSans-Bold.ttf'))
        with self.canvas.before:
            Rectangle(texture=self.img.texture,pos=self.layout.pos,size=self.layout.size)

if __name__ == '__main__':
    # run the app!
    KmpcApp().run()




from setuptools import setup

from kmpc.version import VERSION,VERSION_STR

setup(
    version=VERSION_STR,
    download_url = 'https://github.com/eratosthene/kmpc/archive/'+VERSION_STR+'.tar.gz',
    setup_requires = ['setuptools>=38.4.0'],
    setup_cfg=True,
)
